
/****************************************************************************
 * UGP.C                                                                    *
 ****************************************************************************

Copyright (c) 1998 David Kuehling

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not
   claim that you wrote the original software. If you use this software
   in a product, an acknowledgment in the product documentation would be
   appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be
   misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.

 * Last modification: 20th december, 1998
 *
 * Requires: COMPR92.C
 *
 * This file contains all functions to create UGP and RAW pictures to be
 * displayed on the TI-92 calculator with the assembly shell Fargo (by
 * David Ellsworth).
 *
 * To display UGP pictures, use the library UGPLIB.92p (that was programmed
 * by me), or UGPVIEW.92p (also by me), that can be used to display TIOS
 * strings, that where created by Store_UGP_PIC_to_92S.
 *
 * You should be able to compile this file with any C compiler. UGP.C was
 * originally created using the DJGPP C compiler. However, it would help
 * you very much if you'd compile it with a compiler that is capeable of
 * using DPMI memory.
 *
 * You need to link this against the "packer92.zip" compression algorithm
 * sources.  These should be available on ticalc.org, try
 * https://ticalc.org/archives/files/fileinfo/33/3354.html
 *
 * If you aren't shure how to use special functions, look ugpconv2.c. This
 * file uses all the features of ugp.c.
 */

/**
 *                        The UGP Format
 *                       ----------------
 *
 *   $0000.b:	UGP-version number. (see UGP_VERSION - macro)
 *   $0001.b:	UGP-flags:
 *
 *   Bit:	7	6	5	4	3	2	1	0
 *           ----------------------------------------------------------------
 *          |unused|  slsh  | xor/n | comp  | inter |     planes    |  a/p   |
 *           ----------------------------------------------------------------
 *
 *   a/p:	1 = animation; 0 = picture
 *   planes:	number of planes
 *   inter:	1 = interleaved; 0 = non-interleaved
 *              1 plane pictures mustn't be interleaved. This will crash
 *              UGPLIB. Create_UGP_PIC automatically switches interleaved off,
 *              if planes == 1.
 *   comp:	1 = compressed; 0 = uncompressed
 *   xor/n:	for animations only:
 *	        1 = pictures stored as xor-difference to the last picture,
 *	        0 = pictures stored normaly
 *   slsh:      1 = slide show
 *              0 = animation
 *
 *   $0002.b:	picture's width
 *   $0003.b:	picture's height
 *   $0004.w    if animation: number of pictures
 *              else: begin of picture data (if compressed: PACKER92 -
 * 	              solid archive, containing these data)
 *   $0006:  	begin of animation data (if compressed: normal PACKER92 -
 *              archive, containing all pictures in the right order)
 */


/****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
/****************************************************************************/


/** UGP version number (is stored within the UGP files) */
#define UGP_VERSION 1


/**
 * Import From compr92.c
 ****************************************************************************/
enum ARCTYPES
{
  DIVIDED,
  NORMAL,
  SOLID,
  PROGDATA,
};

extern int              NoOutput;
extern unsigned char    HeaderName[256];
extern void             PackArc (unsigned char *InNm[], unsigned Ins, char *OutNm, enum ARCTYPES ArcType);
/****************************************************************************/


/**
 * Data Structures
 ****************************************************************************/
typedef struct
{
  int           Width, Height;
  unsigned char *Pic;
} BASE_PIC;

typedef struct
{
  int           Width, Height,
                Planes;
  unsigned char *Pic;
} LOWCOLOR_PIC;

typedef struct
{
  int   PicSize;
  char  *Pic;
} RAW_PIC;

typedef struct
{
  struct
  {
    char        UGP_Version;
    unsigned    Animated    : 1;
    unsigned    Planes      : 2;
    unsigned    Interleaved : 1;
    unsigned    Compressed  : 1;
    unsigned    XOR_Ani     : 1;
    unsigned    SlideShow   : 1;
    unsigned    UNUSED      : 1;
    char        Width;
    char        Height;
  } Header;

  char          LowByteOfPicNum;
  char          PicNum;

  unsigned char *Pic;
  int           PicSize;

} UGP_PIC;

typedef struct
{
  int r, g, b;
} LCD_COLOR;

enum DITHERTYPES
{
  NEAREST_SOLID,
  ORD_DITHER1,
  ORD_DITHER2,
  ORD_DITHER3,
  ERROR_DIFF
};
/****************************************************************************/

/**
 * Global Variables
 ****************************************************************************/
enum LOAD_ERROR
{
  NO_ERROR,
  NOT_FOUND,
  TOO_BIG,
  NOT_MUL8,
  BAD_FORMAT
} LoadError;

char *LoadErrorStr[5] =
{
  "no error",
  "file not found",
  "picture is greater than 240x128",
  "picture width is not a multiple of 8",
  "unknown picture format"
};

/** Color constants, for calculating the grey colors */
const LCD_COLOR  TI92_LCD_Black = { 40,  60,  64},
                 TI92_LCD_White = {108, 128,  84};

int ColorNum[] = {0, 2, 4, 7};  /* Number of colors for n-plane picture */

int CompressionOpt;     /* compression optimisation */
/****************************************************************************/



/**
 * Frees the memory, used by a raw picture
 */
void Free_RAW_PIC (RAW_PIC *pic)
{
  free (pic->Pic);
  free (pic);
}

/**
 * Stores a Raw picture into a file. Returns 1 on error, else: 0
 */
int Store_RAW_PIC (RAW_PIC *pic, char *file)
{
  FILE *hdl;
  if ((hdl = fopen (file, "wb")) == NULL)
    return (1);

  if (fwrite (pic->Pic, pic->PicSize, 1, hdl) == -1)
  {
    fclose (hdl);
    return (1);
  }

  fclose (hdl);
  return (0);
}

/**
 * Creates a RAW bitplane picture, from the LOWCOLOR_PIC >src<.
 * No RAW animations are possible. To get animations, you will have to
 * convert every picture manual.
 */
RAW_PIC *Create_RAW_PIC (LOWCOLOR_PIC *pic, int interleaved)
{
  int Ix, Iy,
      Ip,
      bit,
      byte = 0,
      color;
  unsigned char    *src,
                   *dest;
  RAW_PIC *raw;


  raw = malloc (sizeof (RAW_PIC));
  raw->Pic = dest = malloc (raw->PicSize = (pic->Height * pic->Width * pic->Planes) / 8);

  if (interleaved)
    for (Iy = 0, src = pic->Pic, bit = 0; Iy < pic->Width; Iy++)
      for (Ix = 0; Ix < pic->Height; Ix++)
      {
        color = *src++;
        /** if 3 plane picture: 7 greysteps(0..6) -> 0..7 */
        if (pic->Planes == 3 && color > 3)
          color++;

        for (Ip = pic->Planes-1; Ip >= 0; Ip--)
        {
          byte |= (color >> Ip & 1) << bit++;
          if (bit >= 8)
          {
            *dest++ = ~byte;
            bit = byte = 0;
          }
        }
      }
  else
    for (Ip = pic->Planes-1, bit = 7; Ip >= 0; Ip--)
      for (Iy = 0, src = pic->Pic; Iy < pic->Width; Iy++)
        for (Ix = 0; Ix < pic->Height; Ix++)
        {
          color = *src++;
          /** if 3 plane picture: 7 greysteps(0..6) -> 0..7 */
          if (pic->Planes == 3 && color > 3)
            color++;
  
          byte |= (color >> Ip & 1) << bit--;
          if (bit < 0)
          {
            *dest++ = ~byte;
            byte = 0;
            bit = 7;
          }
        }

  return (raw);
}

/**
 * Frees the memory, used by a UGP picture
 */
void Free_UGP_PIC (UGP_PIC *pic)
{
  free (pic->Pic);
  free (pic);
}

/**
 * Stores a UGP - picture into a file. Returns 1 on error, else: 0
 */
int Store_UGP_PIC (UGP_PIC *pic, char *file)
{
  FILE *hdl;
  if ((hdl = fopen (file, "wb")) == NULL)
    return (1);

  if (fwrite (&pic->Header, sizeof (pic->Header), 1, hdl) == -1)
  {
    fclose (hdl);
    return (1);
  }

  if (pic->Header.Animated)
  {
    fputc (pic->LowByteOfPicNum, hdl);
    fputc (pic->PicNum, hdl);
  }

  if (fwrite (pic->Pic, pic->PicSize, 1, hdl) == -1)
  {
    fclose (hdl);
    return (1);
  }

  fclose (hdl);
  return (0);
}

/**
 * Returns the size (in byte) of a UGP picture
 */
#define UGP_SIZE(ugp) ((long)sizeof ((ugp)->Header)+((ugp)->Header.Animated ? 2 : 0) + (ugp)->PicSize)


/**
 * Stores an extended UGP picture, with title, speed and signature into
 * a TI-92 string variable in TI Graph Link format.
 * Returns 1 on error, else: 0
 *
 * >title< mustn't be greater than 30 characters.
 * >anidelay< must be in the range 0-32767
 */
int Store_UGP_PIC_to_92S (UGP_PIC *pic, char *title, int anidelay, char *file)
{
  struct                        /* Graph Link + TIOS file Header */
  {
    char  TI92[8];
    char  Unknown1[2];
    char  Folder[8];
    char  String[42];
    char  Unknown2[4];
    char  Name[8];
    char  Unknown3[4];
    long  GraphLinkLength;
    char  Unknown4[6];
    char  TIOSLength[2];
    char  UGPSignature[4];
  } header;
  unsigned short checksum;      /* Graph Link checksum */
  int      titlelen,            /* length of title */
           varlen,              /* length of the TIOS string variable */
           c;
  char     *name,               /* pointer to name component of filename */
           ename[16],           /* fully extracted name of filename */
           *ext,                /* pointer to file extension */
           *tmp;
  FILE     *hdl;

  if ((hdl = fopen (file, "wb+")) == NULL)
    return (1);

  /* varlen: 4byte UGP signature
   *          + length of title + 1byte trailing '\0'
   *          + odd stringlength(including trailing '\0') ? 1 : 0
   *          + 2byte animation delay time
   *          + size of UGP pic/ani
   *          + 1byte 0 to make TIOS display the string as ""
   *          + 2byte TIOS string suffix (0x002D)
   */
  varlen = (titlelen = strlen (title)) + UGP_SIZE (pic) + ((titlelen&1) ? 10 : 11);

  memcpy (header.TI92, "**TI92**", 8);
  header.Unknown1[0] = 1;
  header.Unknown1[1] = 0;
  memcpy (header.Folder, "main\x0\x0\x0\x0", 8);
  sprintf (header.String, "UGP: %-35s\x1", title);
  header.Unknown2[0] = 0x52;           /* TI-Graph-Link string signature ??*/
  header.Unknown2[1] = 0x00;
  header.Unknown2[2] = 0x00;
  header.Unknown2[3] = 0x00;

  /* get the name part of the filename (without fnsplit) */
  name = (tmp = strchr (file, ':'))  != NULL ? tmp+1 : file;
  name = (tmp = strrchr (file, '/')) != NULL ? tmp+1 : name;
  name = (tmp = strrchr (file, '\\')) >= name ? tmp+1 : name;
  strcpy (ename, name);
  if ((ext = strchr (ename, '.')) != NULL) /* remove extension, if exists */
    *ext = '\0';

  strncpy (header.Name, ename, 8);

  header.Unknown3[0] = 0x0C;
  header.Unknown3[1] = 0x00;
  header.Unknown3[2] = 0x00;
  header.Unknown3[3] = 0x00;

  header.GraphLinkLength = varlen + 0x5A;
  header.Unknown4[0] = 0xA5;
  header.Unknown4[1] = 0x5A;
  header.Unknown4[2] = 0x00;
  header.Unknown4[3] = 0x00;
  header.Unknown4[4] = 0x00;
  header.Unknown4[5] = 0x00;
  header.TIOSLength[0] = varlen>>8;
  header.TIOSLength[1] = varlen&0xFF;
  header.UGPSignature[0] = 'U';
  header.UGPSignature[1] = 'G';
  header.UGPSignature[2] = 'P';
  header.UGPSignature[3] = '\0';

  /** write header and title (+ trailing 0) */
  if (fwrite (&header, sizeof (header), 1, hdl) == -1 ||
      fwrite (title, strlen (title)+1, 1, hdl) == -1)
  {
    fclose (hdl);
    return (1);
  }

  /** seek to next even address */
  if (ftell (hdl)&1)
    fputc (0, hdl);

  /** write animation delay time */
  fputc (anidelay>>8, hdl);
  fputc (anidelay&0xFF, hdl);

  /** write UGP picture */
  if (fwrite (&pic->Header, sizeof (pic->Header), 1, hdl) == -1)
  {
    fclose (hdl);
    return (1);
  }
  if (pic->Header.Animated)
  {
    fputc (pic->LowByteOfPicNum, hdl);
    fputc (pic->PicNum, hdl);
  }
  if (fwrite (pic->Pic, pic->PicSize, 1, hdl) == -1)
  {
    fclose (hdl);
    return (1);
  }

  /** write TIOS string suffix and string end */
  fputc (0x00, hdl);
  fputc (0x00, hdl);
  fputc (0x2D, hdl);

  /** generate checksum and put it to the end of file */
  fseek (hdl, 0x56, SEEK_SET);
  for (checksum = 0, c = fgetc (hdl); !feof (hdl); c = fgetc (hdl))
    checksum += c;

  fseek (hdl, 0, SEEK_END);
  fputc (checksum&0xFF, hdl);
  fputc (checksum>>8, hdl);

  fclose (hdl);
  return (0);
}


/**
 * Creates an UGP animation if picnum > 1, else it creates a picture.
 * >pic< is expected to be an array of pointers to LOWCOLOR_PIC structures.
 * (e.g. LOWCOLOR_PIC *pic[256]; <...> Create_UGP_PIC (pic, .....) )
 * It is expected, that all pictures have the same size and number of planes.
 * You may not use more than 255 pictures.
 * It returns the pointer to the newly allocated UGP_PIC structure, or NULL
 * on error.
 * If compressed=1, it also sets the global variable CompressionOpt to the
 * compression optimisation.
 */
UGP_PIC *Create_UGP_PIC (LOWCOLOR_PIC *pic[], int picnum, int compressed, int interleaved, int xorani, int slideshow)
{
  int     I, J;
  UGP_PIC *ugp;         /* pointer to the ugp picuture, that is created */
  RAW_PIC *raw[255];    /* pointer raw pictures, by that the ugp pic is created */

  /** noninterleaved and interleaved pictures are the same in 1plane mode */
  if (pic[0]->Planes == 1)
    interleaved = 0;

  ugp = malloc (sizeof (UGP_PIC));

  ugp->Header.UGP_Version = UGP_VERSION;
  ugp->Header.Animated    = picnum > 1;
  ugp->Header.Planes      = pic[0]->Planes;
  ugp->Header.Interleaved = interleaved;
  ugp->Header.Compressed  = compressed;
  ugp->Header.XOR_Ani     = xorani;
  ugp->Header.SlideShow   = slideshow;
  ugp->Header.UNUSED      = 0;
  ugp->Header.Width       = pic[0]->Width;
  ugp->Header.Height      = pic[0]->Height;
  ugp->LowByteOfPicNum    = 0;
  ugp->PicNum             = picnum;

  /** convert the pictures to raw pictures */
  for (I = 0; I < picnum; I++)
    raw[I] = Create_RAW_PIC (pic[I], interleaved);

  /** if they should be stored XOR to the last pic -> xor them... */
  if (xorani)
  {
    char    *Ipic1,
            *Ipic2,
            *lastpic,
            byte;

    bzero (lastpic = malloc (raw[0]->PicSize), raw[0]->PicSize);
    for (I = 0; I < picnum; I++)
    {
      for (Ipic1 = raw[I]->Pic, Ipic2 = lastpic, J = 0; J < raw[I]->PicSize; J++, Ipic1++, Ipic2++)
      {
        byte = *Ipic1;
        *Ipic1 ^= *Ipic2;
        *Ipic2 = byte;
      }
    }
    free (lastpic);
  }

  if (compressed)  /* compress all raw pictures into a NORMAL Packer92 archive */
  {
    unsigned char *rawfile[256],     /* names of raw files */
                  *comprfile;        /* name of compressed file */
    long          origsize;          /* original size of pictures */
    FILE *hdl;

    NoOutput = 1;     /* PackArc shouldn't output any information */

    /** store all the pictures into temporary files */
    for (I = 0, origsize = 0; I < picnum; I++)
    {
      rawfile[I] = strdup (tmpnam (NULL));
      if (Store_RAW_PIC (raw[I], rawfile[I]))
        return (NULL);
      origsize += raw[I]->PicSize;
    }

    /** compress them to >comprfile< */
    comprfile = strdup (tmpnam (NULL));
    if (picnum == 1)
      PackArc (rawfile, picnum, comprfile, SOLID);
    else
      PackArc (rawfile, picnum, comprfile, NORMAL);

    /** remove temporary files and the free memory, used by the filenames */
    for (I = 0; I < picnum; I++)
    {
      remove (rawfile[I]);
      free (rawfile[I]);
    }
    remove (HeaderName);

    /** read the compressed file into the ugp - picture */
    if ((hdl = fopen (comprfile, "rb")) == NULL)
      return (NULL);
    ugp->Pic = malloc (ugp->PicSize = filelength (fileno (hdl)));
    if (fread (ugp->Pic, ugp->PicSize, 1, hdl) != 1)
      return (NULL);

    /** calculate the compression optimisation */
    CompressionOpt = ugp->PicSize * 100 / origsize;

    /** remove the compressed file */
    remove (comprfile);
    free (comprfile);
  }
  else /* uncompressed: store all raw pictures, after each other */
  {
    ugp->Pic = malloc (ugp->PicSize = raw[0]->PicSize * picnum);
    for (I = 0; I < picnum; I++)
      memcpy (&ugp->Pic[raw[0]->PicSize * I], raw[I]->Pic, raw[0]->PicSize);
  }

  /** free the raw pictures */
  for (I = 0; I < picnum; I++)
    Free_RAW_PIC (raw[0]);

  return (ugp);
}

/**
 * Converts the pixel (x|y) with the color >color< to the greymode, specified
 * by >planes<. >dithertype< specifies, who colors should be converted.
 */
int PixelColorConv (int x, int y, int dithertype, int planes, int color)
{
  /** tables for dithering */
  static unsigned char ditherdat[4][4][4] =
      {
        {{128, 128, 128, 128}, /* nearest solid */
         {128, 128, 128, 128},
         {128, 128, 128, 128},
         {128, 128, 128, 128}},
        {{ 64, 192,  64, 192}, /* ordered dither 1  - small pattern */
         {192,  64, 192,  64},
         { 64, 192,  64, 192},
         {192,  64, 192,  64}},
        {{224,  96, 224,  96}, /* ordered dither 2  - bigger pattern */
         { 32, 160,  32, 160},
         {224,  96, 224,  96},
         { 32, 160,  32, 160}},
        {{176,  48, 144,  16}, /* ordered dither 3  - huge pattern */
         {112, 240,  80, 208},
         {144,  16, 176,  48},
         { 80, 208, 112, 240}}
      },
      randdat[853] =           /* error diffusion */
      {                    /* ( randomizers are sometimes very slow... ;-) ) */
        189,  2,155,190,127, 81,150, 11,207,158,219, 43, 97,240,111, 15,235, 90, 56,
        182, 72,  1,  2,164,191,137,169,155,166,187, 92, 99,189,248, 34, 61, 73,184,
         72, 24, 86, 35, 67,183, 19,178,199,255, 12,255,181, 84,  0,184,248,191, 65,
        162, 91,231, 93,183, 75, 27,175,109, 88,248, 37,160, 17,124,196, 84, 51,215,
          7,250,214, 19,249,140,104,250, 68, 96,185,133,  2, 20,109, 96,204,184,123,
        123, 37,211,116, 74,115,133,198, 55,217,250, 15,224,244,229,244,238,113, 92,
        232,181,188,161, 59,191,182,168, 31,130, 96,154,253,133,109,113,207,224,246,
        150, 24,208,144, 39,176,132, 12,164,114,126,  0, 90, 51,189,252,110,124,178,
         22,155, 52,118, 53, 49,251,162,163,203,130,153, 97,154,105,241,193, 26,117,
        206,190,232, 76,191, 66,127,124, 62,238,248,240,  4,147, 36,123,200, 86,118,
        106,249, 65,236,146,162,135,252,147, 72, 22,  9, 22,212,241, 98,147, 51,226,
         15,114,208,  7, 98,212,154,135, 79, 98,221,198,204,214,  7,185,104,170, 64,
        100, 61,136,122, 70,159, 79, 55,  1,226,107,227,242,221,179,249, 63,136,148,
        198,215,246,163,157,195,121,165,124,226, 79,188, 70,140, 68,193,211,227, 16,
         10,229,242,117,200,228, 82,124,222,146,  4,114, 88,219,104,252,121, 43,117,
         30,167, 87,109, 99,158,249,168, 95,204,139,111,215,112, 97, 76, 57, 70,159,
        181, 36, 49,185,150,137,148,254,133, 13, 42,251, 43,209, 82,152, 53,240,146,
        221, 79, 94,104,190, 53,217, 32,130, 18,102, 33,199,138, 82,128, 32,219, 20,
         30, 97, 34, 72, 92, 77, 26,174,230, 79,159,120, 44,238,214,148,173, 12,109,
        205,142,127, 51,175, 70,189,  1,198,221,220,219,251, 61,253, 68,153, 74, 94,
         72, 48,173,231,168,217,213,127,109,130,139,219, 79, 25, 90,130,200,161, 63,
        201,103, 28,165, 66, 24,227, 63, 92,124,138,186,196,186,103,171, 99, 64,129,
        226,173,  3,109,136, 83,134,227,213, 78,132, 21, 23,235, 49,188, 46, 73,159,
        109,165, 28,247, 95,224,178,198,140, 21,  6, 13,247,180, 16,100, 60, 99,234,
         31, 57, 56,163, 78, 79,143,127, 11,189,201,171, 42,110,199, 34,206,167,212,
        148, 51,233,155, 64,224, 79, 81, 68,139,180, 46,171,237,102, 78, 59,181,221,
        187,192,154,132,107,197,242, 50,231,192,218,187, 85, 13,164,240, 78,132, 63,
        159,200,202, 83,246,117, 65, 92,196,124, 17,161, 55,209, 60,187, 61,  1,174,
        111,232,110, 73,163,195, 87, 71,179,165,203,242, 68,147,189,151,137, 50,216,
        229,246, 85,246,152,140,199,212, 72,  4,213,246,116,189,100,189, 96, 40, 20,
        167,219,185,114,206,253,  5,139,149,142,189,109,115,180,194,105, 76, 79, 48,
         32,151, 53,245,141,169,178,241,102, 18, 25,123,185,245, 52, 43,195, 50, 48,
         78,199,190, 11, 52, 49,191,247,154, 11, 70,202, 43,221,255, 32,106,168,210,
         91, 15,228,117,138,157,106,190,200, 45,240,248,123,183,182,134,236,231, 70,
        227,129, 81, 41, 76,125,  6, 75,157,112,244,112,203,  3, 84, 64,141,242,170,
         75,186,215, 60,179, 82,243,105,217,223, 81, 31,194,210,112,235, 30,237,241,
        106,139, 97, 94,251, 45, 97, 79,109,238, 65, 24, 57,252,239,117,175, 66,105,
         24, 27, 72,105, 58, 11, 60,170,246, 90,152,232,196, 35, 73, 34, 30,118,131,
        109,228,113,175,252,171,171,235, 32, 90, 45,137,114, 72,210,220,130,221, 24,
         45,211,114,197,187, 55,232,  5, 89,  6,123,221,115, 95, 78, 34, 91,249,205,
         71, 26, 39,116,163,154,189,117,118, 63, 82,142,108, 38,  0, 49,225, 55, 25,
        230,145, 31, 98,110,147,193,188,181, 29,182,131,100,208,170,216,115, 68,149,
        233,186,213, 59, 72, 65, 97, 73,115, 67,128,140, 41, 17,172,139,127, 63, 77,
         60,244,106,242,119,206,194, 34,166, 53,102, 60, 30, 33, 17, 90,105, 82,187,
        178,197,254, 51, 82, 40, 68,254,179,196, 61,  0,  0, 49,106,242,169
      };
  int maincol,
      remain;

  maincol = color*(ColorNum[planes] - 1) >> 8;
  remain  = color*(ColorNum[planes] - 1) & 255;

  if (remain >= (dithertype == ERROR_DIFF ?
                   randdat[(y*113 + x)%853] :
                   ditherdat[dithertype][x&3][y&3]))
    return (maincol+1);

  return (maincol);
}

/**
 * Frees the memory, used by the LOWCOLOR_PIC pic.
 */
void Free_LOWCOLOR_PIC (LOWCOLOR_PIC *pic)
{
  free (pic->Pic);
  free (pic);
}

/**
 * Reduces the color of the base picture (BASE_PIC), to the number of planes,
 * and returns a newly allocated LOWCOLOR_PIC structure.
 * Contrast: 0..255 (normal = 128)
 * Brightness: 0..255 (normal = 128)
 */
LOWCOLOR_PIC *ColorConv (BASE_PIC *base, int brightness, int contrast, int planes, int dithering)
{
  LOWCOLOR_PIC  *low;          /* the new >planes< bitplane picture */
  int           Ix, Iy,
                color;
  unsigned char *dest,
                *src;


  low = malloc (sizeof (LOWCOLOR_PIC));
  low->Planes = planes;
  low->Width = base->Width;
  low->Height = base->Height;
  low->Pic = dest = malloc (base->Width*base->Height);
  src = base->Pic;

  srandom (0); /* random dithering should allways look the same */

  for (Iy = 0; Iy < base->Height; Iy++)
    for (Ix = 0; Ix < base->Width; Ix++)
    {
      color = *src++;

      color = ((color-128) * contrast >> 7) + 128;
      color += brightness-128;
      color = color < 0 ? 0 : color > 255 ? 255 : color;

      *dest++ = PixelColorConv (Ix, Iy, dithering, planes, color);
    }

  return (low);
}

/**
 * Calculates the red, green and blue component of a color from a
 * LOWCOLOR_PIC.Pic picture, as it will be displayed on the TI-92 LCD
 * with the number of specified planes. The color-values will be in
 * a range between 0..255.
 **
 * Return:
 *  Pointer to the newly allocated LCD_COLOR structure
 */
LCD_COLOR *getColorRGB (int color, int planes)
{
  LCD_COLOR *rgb;
  rgb = malloc (sizeof (LCD_COLOR));

  rgb->r = TI92_LCD_Black.r + color*(TI92_LCD_White.r - TI92_LCD_Black.r) / (ColorNum[planes]-1);
  rgb->g = TI92_LCD_Black.g + color*(TI92_LCD_White.g - TI92_LCD_Black.g) / (ColorNum[planes]-1);
  rgb->b = TI92_LCD_Black.b + color*(TI92_LCD_White.b - TI92_LCD_Black.b) / (ColorNum[planes]-1);

  return (rgb);
}

/**
 * Frees the memory, used by the LCD_COLOR structure
 */
void Free_LCD_COLOR (LCD_COLOR *rgb)
{
  free (rgb);
}

/**
 * Frees the memory, used by a BASE_PIC structure.
 */
void Free_BASE_PIC (BASE_PIC *pic)
{
  free (pic->Pic);
  free (pic);
}

/**
 * Loads a 24bit, uncompressed .TGA file, and stores it into a BASE_PIC
 * structure. The memory, allocated by this routine can be freed by
 * Free_BASE_PIC.
 * Returns a pointer to the created BASE_PIC structure, or NULL on error.
 * If an error occures, it sets the LoadError variable.
 */
BASE_PIC *Load_TGA_24bit_uncompressed (char *file)
{
  BASE_PIC *pic;                /* picture that is returned */
  struct {                      /* data, read from the .tga - file */
    char      Unknown1[12];
    short     Width,
              Height;
    char      Unknown2,
              TopBottom;
  } header;
  long     size,                /* Number of pixels in picture */
           I, J;
  char     line[240];
  FILE     *hdl;

  if ((hdl = fopen (file, "rb")) == NULL)
  {
    LoadError = NOT_FOUND;
    return (NULL);
  }

  pic = malloc (sizeof (BASE_PIC));

  if (fread (&header, sizeof (header), 1, hdl) != 1 ||
      memcmp (header.Unknown1, "\x0\x0\x2\x0\x0\x0\x0\x0\x0\x0\x0\x0", 12))
  {
    LoadError = BAD_FORMAT;
    return (NULL);
  }

  if (header.Width > 240 || header.Height > 128)
  {
    LoadError = TOO_BIG;
    return (NULL);
  }

  if (header.Width&7)          /* Picture width is not a multiple of 8 ? */
  {
    LoadError = NOT_MUL8;
    return (NULL);
  }
  pic->Width = header.Width;
  pic->Height = header.Height;

  /** read the picture data (24bit red,green,blue -> 8bit grey) */
  pic->Pic = malloc (size = header.Width * header.Height);
  for (I = 0; I < size; I++)
  {
    pic->Pic[I] = (fgetc (hdl) + fgetc (hdl) + fgetc (hdl)) / 3;
    if (feof (hdl))
    {
      LoadError = BAD_FORMAT;
      return (NULL);
    }
  }

  /** picture is stored Bottom-->Top -> has to be line-reversed */
  if (!header.TopBottom)
  {
    for (I = 0, J = header.Height-1; I < header.Height/2; I++, J--)
    {
      /* exchange two lines */
      memcpy (line, &pic->Pic[I*header.Width], header.Width);
      memcpy (&pic->Pic[I*header.Width], &pic->Pic[J*header.Width], header.Width);
      memcpy (&pic->Pic[J*header.Width], line, header.Width);
    }
  }

  return (pic);
}

