/****************************************************************************
 * UGP.C                                                                    *
 ****************************************************************************

Copyright (c) 1998 David Kuehling

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not
   claim that you wrote the original software. If you use this software
   in a product, an acknowledgment in the product documentation would be
   appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be
   misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.

 * This is the header file of the UGP.C graphic conversion functions. (you
 * will also have to link COMPR92.C to it)
 * For further information look UGP.C.
 *
 */


/**
 * How it works:
 *
 * 1. Load picture(s) with one of the load functions (currently only one
 *    routine for loading 24bit uncompressed TGA files is implemented).
 *    Pictures will be loaded into a BASE_PIC structure. This structure
 *    contains the graphic(s) grey converted with 8 bit/pixel, it will be
 *    the base for color reducing etc. and will allways stay the same.
 * 2. Convert the picture(s) with the >ColorConv< function to a picture with
 *    a certain number of planes. This structure contains the picture in
 *    a format, that can easyly be displayed, 1byte/pixel. 0 is black, and
 *    the number of white will depend on the number of planes you specified.
 *    To get out the color of a certain color-number, as it will be on the
 *    TI-92's LCD display, use >getColorRGB<. Note that the 3 plane mode, as
 *    supported by gray7lib, in fact only displays 7 grey-colors, thus color
 *    #3 and #4 are the same (this was propably done to reduce flickering).
 *    The convertion can be done many times, with different brightness and
 *    contrast settings, to let the user try what looks best. Allways
 *    remember, that you have to free the LOWCOLOR_PIC pictures, before
 *    reusing them.
 * 3. The LOWCOLOR_PIC picture(s), you gained by (2.) can now be converted
 *    into raw picture(s) or into one UGP picture, by >Create_RAW_PIC< or
 *    >Create_UGP_PIC<. If you pass more than one LOWCOLOR_PIC structure to
 *    >Create_UGP_PIC<, a UGP animation will be created else the result is a
 *    UGP picture. Raw pictures can only contain _one_ picture. If you want
 *    to create an animation of raw pictures, they have to be stored into
 *    seperated files.
 * 4. Store the UGP/RAW picture into a file, by Store_UGP_PIC or
 *    Store_RAW_PIC. UGP pictures can also be stored into a TI Graph Link
 *    string variable (*.92s), by Store_UGP_PIC_to_92S. Such string
 *    variables can be displayed by UGPVIEW.92p or UGPBRWSE.92p
 *
 * However, you should look up ugpconv2.c, it uses all the features that are
 * supported by ugp.c.
 *
 */

#ifndef __UGP_H
#define __UGP_H

/**** data structures **/
typedef struct
{
  int   Width, Height;
  char  *Pic;
} BASE_PIC;

typedef struct
{
  int   Width, Height,
        Planes;
  char  *Pic;
} LOWCOLOR_PIC;

typedef struct
{
  int   PicSize;
  char  *Pic;
} RAW_PIC;

typedef struct
{
  struct
  {
    char        UGP_Version;
    unsigned    Animated    : 1;
    unsigned    Planes      : 2;
    unsigned    Interleaved : 1;
    unsigned    Compressed  : 1;
    unsigned    XOR_Ani     : 1;
    unsigned    UNUSED      : 2;
    char        Width;
    char        Height;
  } Header;

  char          LowByteOfPicNum;
  char          PicNum;

  char          *Pic;
  int           PicSize;

} UGP_PIC;

typedef struct
{
  int r, g, b;
} LCD_COLOR;

/* dithering types */
enum DITHERTYPES
{
  NEAREST_SOLID,
  ORD_DITHER1,
  ORD_DITHER2,
  ORD_DITHER3,
  ERROR_DIFF
};

/* global error variable, is set by the picture-load functions */
extern enum LOAD_ERROR
{
  NO_ERROR,
  NOT_FOUND,      /* file not found */
  TOO_BIG,        /* picture greater than 240x128 */
  NOT_MUL8,       /* picture width is not a multiple of 8 */
  BAD_FORMAT      /* unknown picture format */
} LoadError;

/**
 * strings, that describe every error, indicated by LoadError
 * (e.g. printf ("Load Error: %s\n", LoadErrorStr[LoadError]);
 */
extern char *LoadErrorStr[5];

/**
 * Number of colors for n-plane picture.
 * (e.g. ColorNum[pic->Planes])
 */
extern int ColorNum[],
           CompressionOpt;

/**
 * Black and white, as they are displayed on the TI-92 display.
 */
extern const LCD_COLOR  TI92_LCD_Black,
                        TI92_LCD_White;

/**** function declarations **/
#ifdef __cplusplus
extern "C" {
#endif

/**
 * Loads a 24bit, uncompressed .TGA file, and stores it into a BASE_PIC
 * structure. The memory, allocated by this function can be freed by
 * Free_BASE_PIC (BASE_PIC *pic)
 **
 * Return:
 *  Pointer to the created BASE_PIC structure, or NULL on error.
 *  If an error occures, it sets the global LoadError variable.
 */
BASE_PIC *Load_TGA_24bit_uncompressed (char *file);

/**
 * Frees the memory, used by a BASE_PIC structure.
 */
void Free_BASE_PIC (BASE_PIC *pic);

/**
 * Reduces the color of the base picture (BASE_PIC) to the number of
 * specified bitplanes, and returns a newly allocated LOWCOLOR_PIC structure.
 * The LOWCOLOR_PIC structure contains all the pixels of the picture as an
 * array of chars. Pixels can be accessed seperated by the PIXEL(pic,x,y) -
 * macro
 **
 * Contrast:   0..255 (normal = 128, 0 = low, 255 = very high)
 * Brightness: 0..255 (normal = 128, 0 = dark, 255 = very bright)
 * Dithering:  (enum DITHERTYPES)
 *             NEAREST_SOLID - take the nearest matching solid color
 *             ORD_DITHER1   - small pattern dithering
 *             ORD_DITHER2   - big pattern dithering
 *             ERROR_DIFF    - error diffusion dithering (no pattern)
 **
 * Return:
 *   Pointer to the created BASE_PIC structure
 */
LOWCOLOR_PIC *ColorConv (BASE_PIC *base, int brightness, int contrast, int planes, int dithering);

/**
 * Frees the memory, used by the LOWCOLOR_PIC structure.
 */
void Free_LOWCOLOR_PIC (LOWCOLOR_PIC *pic);

/**
 * This macro returns the color of the pixel (x|y) of the LOWCOLOR_PIC picture
 * >pic<. The range of the color depends on the number of planes, of the
 * picture.
 */
#define PIXEL(pic,x,y) ((pic)->Pic[(y) * (pic)->Width + (x)])

/**
 * Calculates the red, green and blue component of a color from a
 * LOWCOLOR_PIC.Pic picture, as it will be displayed on the TI-92 LCD
 * with the number of specified planes. The color-values will be in
 * a range between 0..255.
 **
 * Return:
 *  Pointer to the newly allocated LCD_COLOR structure
 */
LCD_COLOR *getColorRGB (int color, int planes);

/**
 * Frees the memory, used by the LCD_COLOR structure
 */
void Free_LCD_COLOR (LCD_COLOR *rgb);

/**
 * Creates an UGP animation if picnum > 1, else it creates a picture.
 * >pic< is expected to be an array of pointers to LOWCOLOR_PIC structures.
 * (e.g. LOWCOLOR_PIC *pic[256]; <...> Create_UGP_PIC (pic, .....) )
 *
 * All pictures must have the same size and number of planes. The maximum
 * numbers of pictures is 255.
 **
 * picnum:      number of pictures (1=create UGP-picture instead of animation)
 * compressed:  != 0: compress picture(s)
 *                    the compression optimisation is stored into the global
 *                    variable CompressionOpt
 * interleaved: != 0: store pictures interleaved (all planes merged together)
 * xorani:      != 0: store animation as XOR animation (each picture is the
 *                    XOR - difference to the last picture)
 * slideshow:   != 0: mark UGP picture to be displayed as slide show instead
 *                    animation. Pictures are displayed until key is pressed.
 **
 * Return:
 *  Pointer to the newly allocated UGP_PIC structure, or NULL on error.
 **
 * Note, that it can take quite a while, to convert pictures, if they should
 * be compressed, due to the complexity of the compression algorithm.
 */
UGP_PIC *Create_UGP_PIC (LOWCOLOR_PIC *pic[], int picnum, int compressed, int interleaved, int xorani, int slideshow);

/**
 * Frees the memory, used by a UGP_PIC structure.
 */
void Free_UGP_PIC (UGP_PIC *pic);

/**
 * Stores a UGP - picture into a file.
 **
 * Return:
 *    no error: 0
 *    error: 1 this should allways be a file error - strerror (errno) or
 *           perror ("...") could output an error message
 */
int Store_UGP_PIC (UGP_PIC *pic, char *file);

/**
 * Stores an extended UGP picture, with title, speed and signature into
 * a TI-92 string variable in TI Graph Link format.
 * Returns 1 on error, else: 0
 **
 * title:    title of UGP picture, as it will be displayed by UGPVIEW -
 *           mustn't be greater than 30 characters
 * anidelay: animation: delay time per picture (1 = 1/350 secs) - must be in
 *           the range 0-32767
 **
 * Return:
 *    no error: 0
 *    error: 1 this should allways be a file error - strerror (errno) or
 *           perror ("...") could output an error message.
 */
int Store_UGP_PIC_to_92S (UGP_PIC *pic, char *title, int anidelay, char *file);


/**
 * Calculates the total size of a UGP picture in bytes.
 * Note, that it returns a long value in case you want to output it with
 * printf etc.
 */
#define UGP_SIZE(ugp) ((long)sizeof ((ugp)->Header)+((ugp)->Header.Animated ? 2 : 0) + (ugp)->PicSize)

/**
 * Creates one RAW bitplane picture, from the LOWCOLOR_PIC >src<.
 **
 * interleaved: !=0: store picture interleaved (all planes merged together)
 **
 * Return:
 *  Pointer to the newly created RAW_PIC structure.
 */
RAW_PIC *Create_RAW_PIC (LOWCOLOR_PIC *pic, int interleaved);

/**
 * Frees the memory, used by a RAW_PIC structure
 */
void Free_RAW_PIC (RAW_PIC *pic);

/**
 * Calculates the size of a raw picture in bytes.
 */
#define RAW_SIZE(pic) ((pic)->PicSize)

/**
 * Stores a Raw picture into a file.
 * Return:
 *    no error: 0
 *    error: 1 this should allways be a file error - strerror (errno) or
 *           perror ("...") could output an error message
 */
int Store_RAW_PIC (RAW_PIC *pic, char *file);

#ifdef __cplusplus
}
#endif

#endif /* __UGP_H */
